"""
the new models will replace of the old <<model>>
======
ref list:
TensorFlow2.0教程-AutoGraph https://zhuanlan.zhihu.com/p/59482934
Writing a new Tensorflow operation (including C++, CUDA, forward, gradient and grad check) https://medium.com/@taxfromdk/writing-a-new-tensorflow-operation-including-c-cuda-forward-gradient-and-grad-check-3c46708351e7
Get Gradients with Keras Tensorflow 2.0 https://stackoverflow.com/questions/57759635/get-gradients-with-keras-tensorflow-2-0
tf.function and AutoGraph in TensorFlow 2.0 https://www.kaggle.com/vikramtiwari/tf2-tf-function-and-autograph
Writing custom layers and models with Keras https://www.tensorflow.org/guide/keras/custom_layers_and_models
https://medium.com/@chinesh4/custom-activation-function-in-tensorflow-for-deep-neural-networks-from-scratch-tutorial-b12e00652e24
https://datascience.stackexchange.com/questions/58884/how-to-create-custom-activation-functions-in-keras-tensorflow
https://stackoverflow.com/questions/37657260/how-to-implement-custom-metric-in-keras
Tensorflow 中的损失函数 —— loss 专题汇总 https://zhuanlan.zhihu.com/p/44216830
定义你自己的Tensorflow Optimizer https://github.com/luochuwei/Custom-Optimizer-in-TensorFlow
1.1. A ~5 minute guide to Numba https://numba.pydata.org/numba-doc/dev/user/5minguide.html
"""