"""
因为support tensorflow2.0的models现在还没有pip安装包，故从github下载了相关的代码
https://github.com/tensorflow/models/blob/master/official/nlp/bert_modeling.py
unzip: https://codeload.github.com/tensorflow/models/zip/v2.1.0/models-2.1.0.zip
@date: 2020-02-05
@since TensorFlow Official Models 2.1.0
支持tensorflow2.0 for bert
"""