"""
https://github.com/openai/baselines/tree/tf2
https://github.com/hill-a/stable-baselines
https://www.reddit.com/r/algotrading/comments/dxcnaq/deep_reinforcement_learning_portfolio_management
https://www.slideshare.net/KamerAliYuksel/deep-reinforcement-learning-portfolio-management
"""